/*	$Id: kurvmatch.c,v 1.15 2013/12/31 04:33:09 ragge Exp $	*/

#include <err.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "kurvmatch.h"

#include "gsl/gsl_multimin.h"

#define NPARAM 6
/*
 * These elements are overlayed for different types of tubes:
 *		1	2	3	4	5	6
 * Diodes:	KA	EX	KB	-
 * Triodes:	MU	EX	KG1	KP	KVB	
 */

struct tudata d;
int dflag, sflag, gflag, tflag, uflag, hflag;

/*
 * Calculation routine.	 
 * KA, EX, KB, EPS
 */
double
diode(const gsl_vector *v, void *params)
{
	double ka, kb, ex, vct, eps;
	struct tudata *d = params;
	
	double y;
	int i;

	ka = gsl_vector_get(v, xKA);
	ex = gsl_vector_get(v, xEX);
	kb = gsl_vector_get(v, xKB);
	eps = d->EPS;
	y = 0.0;
	vct = d->VCT;

	for (i = 0; i < d->n; i++) {
		double vp, ip, y1;

		vp = d->Vp[i];
		ip = d->Ip[i];

		y1 = vp >= 0 ? (ka+kb*vp) * pow(vp+eps, ex) : 0;
		y1 -= ip;

		y += y1 * y1;

	}
	y = sqrt(y);
	return y;
}

/*
 * Calculation routine.	 
 * MU, EX, KG1, KP, KVB
 */
double
pentode(const gsl_vector *v, void *params)
{
	double mu, ex, kg1, kp, kvb, vct;
	struct tudata *d = params;
	
	double y;
	int i;

	mu = gsl_vector_get(v, xMU);
	ex = gsl_vector_get(v, xEX);
	kg1 = gsl_vector_get(v, xKG1);
	kp = gsl_vector_get(v, xKP);
	kvb = gsl_vector_get(v, xKVB);
	y = 0.0;
	vct = d->VCT;

	for (i = 0; i < d->n; i++) {
		double vp, vg, vg2, ip, E1, y1;

		vp = d->Vp[i];
		vg = d->Vg[i];
		vg2 = d->Vg2[i];
		ip = d->Ip[i];

		E1 = (vg2/kp) * log(1+exp(kp*(1/mu+(vg+vct)/vg2)));
		y1 = E1 >= 0 ? (pow(E1, ex)/kg1) * atan(vp/kvb) : 0;
		y1 -= ip;

		y += y1 * y1;

	}
	y = sqrt(y);
	return y;
}

/*
 * Calculation routine.	 
 * MU, EX, KG1, KP, KVB
 */
double
triode(const gsl_vector *v, void *params)
{
	double mu, ex, kg1, kp, kvb, vct;
	struct tudata *d = params;
	
	double y;
	int i;

	mu = gsl_vector_get(v, xMU);
	ex = gsl_vector_get(v, xEX);
	kg1 = gsl_vector_get(v, xKG1);
	kp = gsl_vector_get(v, xKP);
	kvb = gsl_vector_get(v, xKVB);
	y = 0.0;
	vct = d->VCT;

	for (i = 0; i < d->n; i++) {
		double vp, vg, ip, E1, y1;

		vp = d->Vp[i];
		vg = d->Vg[i];
		ip = d->Ip[i];

		E1 = (vp/kp) * log(1+exp(kp*(1/mu+(vg+vct)/sqrt(kvb+vp*vp))));
		y1 = E1 >= 0 ? pow(E1, ex)/kg1 : 0;
		y1 -= ip;

		y += y1 * y1;

	}
	y = sqrt(y);
	return y;
}

static void
diprint(int n, gsl_multimin_fminimizer *s)
{
	printf("%5d:   KA %.6f EX %.2f KB %.6f f() = %7.3f size = %.3f\n",
	    n,
	    gsl_vector_get (s->x, xKA),
	    gsl_vector_get (s->x, xEX),
	    gsl_vector_get (s->x, xKB),
	    s->fval, gsl_multimin_fminimizer_size (s));
}

static void
triprint(int n, gsl_multimin_fminimizer *s)
{
	printf("%5d:   MU %.0f EX %.2f KG1 %.0f KP %.0f KVB %.0f f() = %7.3f size = %.3f\n",
	    n,
	    gsl_vector_get (s->x, 0),
	    gsl_vector_get (s->x, 1),
	    gsl_vector_get (s->x, 2),
	    gsl_vector_get (s->x, 3),
	    gsl_vector_get (s->x, 4),
	    s->fval, gsl_multimin_fminimizer_size (s));
}

int
main(int argc, char *argv[])
{
	int status, r, i, ch, iter = 0;
	int yyparse (void);
	const gsl_multimin_fminimizer_type *T =
	    gsl_multimin_fminimizer_nmsimplex2;
	gsl_multimin_fminimizer *s = NULL;
	gsl_vector *var, *x;
	gsl_multimin_function f;
	double *dx, *dvar, size, tsize;
	void (*p)(int, gsl_multimin_fminimizer *);

	while ((ch = getopt(argc, argv, "dghstu")) != -1) {
		switch (ch) {
		case 'd':	/* Debug */
			dflag = 1;
			break;
		case 'g':	/* Generate gnuplot curve */
			gflag = 1;
			break;
		case 'h':	/* Print gnuplot curve to .png file */
			hflag++;
			break;
		case 's':	/* Generate spice code */
			sflag = 1;
			break;
		case 't':	/* Generate test code for spice */
			tflag++;
			break;
		case 'u':	/* Append spice code to 'tubelib.sub' */
			uflag++;
			break;
		case '?':
		default:
			errx(1, "Usage: %s [-d] [-s] [-g] [-t] [configfile]", argv[0]);
		}
	}
	argc -= optind;
	argv += optind;


	/* Read tube data.  */
	if (argc == 1 && freopen(argv[0], "r", stdin) == NULL)
		errx(1, "Failed open data file");

	yyparse();

	/* Sanitychecks */
	if (d.name == NULL)
		errx(1, "missing tube name");

	switch (d.type) {
	case DIODE:
		f.n = 3;
		f.f = diode;
		tsize = 1e-5;
		dx = (double []){0.005, 1.4, 0.0002};
		dvar = (double []){0.001, 0.01, 0.0001};
		p = diprint;
		break;

	case TRIODE:
		f.n = 5;
		f.f = triode;
		tsize = 1e-3;
		dx = (double []){50, 1.3, 1000, 200, 100};
		dvar = (double []){1, .01, 1, 1, 1 };
		p = triprint;
		break;

	case PENTODE:
		f.n = 5;
		f.f = pentode;
		tsize = 1e-3;
		dx = (double []){50, 1.3, 1000, 200, 100};
		dvar = (double []){1, .01, 1, 1, 1 };
		p = triprint;
		break;

	default:
		errx(1, "unknown tube type %d", d.type);
	}
	f.params = &d;

	if (d.MU != 0)
		dx[xMU] = d.MU;

	x = gsl_vector_alloc (f.n);
	var = gsl_vector_alloc (f.n);
	for (i = 0; i < f.n; i++) {
		gsl_vector_set (x, i, dx[i]);
		gsl_vector_set (var, i, dvar[i]);
	}

	s = gsl_multimin_fminimizer_alloc (T, f.n);
	gsl_multimin_fminimizer_set (s, &f, x, var);

	printf("----------------------------- Iteration ------------\n");
	do {
		iter++;
		if ((status = gsl_multimin_fminimizer_iterate(s)))
			break;
		size = gsl_multimin_fminimizer_size (s);
		if ((status = gsl_multimin_test_size (size, tsize)) == GSL_SUCCESS)
			printf ("converged to minimum at\n");

		(*p)(iter, s);
	} while (status == GSL_CONTINUE && iter < 5000);

	for (i = 0; i < f.n; i++)
		dx[i] = gsl_vector_get (s->x, i);

	if (dx[xKVB] < 0.0)
		warnx("WARNING: bad matching, KVB negative");

	if (gflag || tflag) {
		char fb[16384];
		double maxp, maxi, maxg, id;
		int mp, mi, mg, mt;
		char fn[] = "/tmp/kurvor.XXXXXX";
		FILE *fd;

		/* get x/y coordinates */
		for (i = 0; i < d.n; i++) {
			if (d.Vp[i] > maxp) maxp = d.Vp[i];
			if (d.Ip[i] > maxi) maxi = d.Ip[i];
			if (d.Vg[i] < maxg) maxg = d.Vg[i];
		}
		mp = maxp/100; mp++; mp *= 100;
		mi = maxi * 1000; mi++; 
		mg = maxg; mt = 0; if (mg > -5) mt = 1; if (mg < -10) mt = 2;

		/* If not given in config file, use above */
		if (d.Xt == 0.0)
			d.Xt = mp, d.Xi = 20.0;
		if (d.Yt == 0.0)
			d.Yt = (double)mi/1000, d.Yi = .0005;
		if (d.Gt == 0.0)
			d.Gt = mg, d.Gi = (mg > -5 ? -.5 : mg < -10 ? -2.0 : -1.0);

		if ((i = mkstemp(fn)) < 0)
			err(1, "mkstemp");
		if ((fd = fdopen(i, "w")) == NULL)
			errx(1, "fdopen");

		switch (d.type) {
		case DIODE:
			fprintf(fd, "Ic(Vp) = (%f+Vp*%f)*(Vp+%f)**%f\n",
			    dx[xKA], dx[xKB], d.EPS, dx[xEX]);
			break;

		case TRIODE:
			fprintf(fd, "E1(Vp,Vg) = "
			    "(Vp/%.1f)*log(1.0+exp(%.1f*(1.0/%.0f+(Vg+%.1f)/sqrt(%.1f+Vp*Vp))))\n",
			    dx[xKP], dx[xKP], dx[xMU], d.VCT, dx[xKVB]);
			fprintf(fd, "Ic(Vp,Vg) = (E1(Vp,Vg)**%.2f)/%.0f\n", dx[xEX], dx[xKG1]);
			break;

		case PENTODE:
			fprintf(fd, "E1(Vp,Vg,Vg2) = "
			    "(Vg2/%.1f)*log(1.0+exp(%.1f*(1.0/%.0f+(Vg+%.1f)/Vg2)))\n",
			    dx[xKP], dx[xKP], dx[xMU], d.VCT);
			fprintf(fd, "Ic(Vp,Vg,Vg2) = ((E1(Vp,Vg,Vg2)**%.2f)/%.1f)*atan(Vp/%f)\n",
			    dx[xEX], dx[xKG1], dx[xKVB]);
			break;
		default:
			errx(1, "Unknown gnuplot tube");
		}
		fprintf(fd, "set grid;set xtics %.1f;set ytics %.4f\n", d.Xi, d.Yi);
		fprintf(fd, "set title \"%s - %s\"\n", d.name, d.descr);

		switch (d.type) {
		case DIODE:
			break;

		case PENTODE:
		case TRIODE:
			for (i = 0; i < d.n; i++) {
				if (d.type != PENTODE || d.Vg2[i] == 250)
					fprintf(fd, "set label \"(%.1f)\" at %.0f,%.4f\n",
					    d.Vg[i], d.Vp[i]+5, d.Ip[i]);
			}
			break;

		default:
			errx(1, "Unknown gnuplot tube2");
		}

		if (hflag) { /* Print to file */
			fprintf(fd, "set terminal png enhanced font \"Courier,8\"\n");
			fprintf(fd, "set output \"%s.png\"\n", d.name);
		}

		fprintf(fd, "plot [%.0f:%.0f] [%.4f:%.4f] ", d.Xb, d.Xt, d.Yb, d.Yt);
		switch (d.type) {
		case DIODE:
			fprintf(fd, "Ic(x),");
			break;
		case TRIODE:
			for (id = d.Gb; id >= d.Gt; id += d.Gi)
				fprintf(fd, "Ic(x,%.1f) title '%.1f',", id, id);
			break;
		case PENTODE:
			for (id = d.Gb; id >= d.Gt; id += d.Gi)
				fprintf(fd, "Ic(x,%.1f,250) title '%.1f',", id, id);
			break;
		default:
			errx(1, "Unknown gnuplot tube3");
		}

		fprintf(fd, "\"<echo '");

		for (i = 0; i < d.n; i++) {
			if (d.type != PENTODE || d.Vg2[i] == 250)
				fprintf(fd, "%.0f %.4f\\n", d.Vp[i], d.Ip[i]);
		}
		fprintf(fd, "'\" with points lt 1 notitle");
		if (d.Wa)
			fprintf(fd, ",%.2f/x title 'Wa'", d.Wa);
		fprintf(fd, "\n");
		fclose(fd);

		printf("\n----------------------------- GNUplot args ------------\n");
		sprintf(fb, "cat %s", fn);
		system(fb);
		sprintf(fb, "gnuplot %s", fn);
		system(fb);
		unlink(fn);

		if (tflag) {
			/* Generate SPICE test code */
			printf("\n----------------------------- SPICE test code ------------\n");
			printf("* %s - %s test code\n", d.name, d.descr);
			switch (d.type) {
			case DIODE:
				printf("XU1 Vp 0 %s\n", d.name);
				printf("V2 Vp 0 10\n");
				printf(".dc V2 %.0f %.0f %.0f\n", d.Xb, d.Xt, d.Xi);
				break;
			case TRIODE:
				printf("XU1 Vp Vg 0 %s\n", d.name);
				printf("V2 Vp 0 100\n");
				printf("V1 0 Vg 0\n");
				printf(".dc V2 %.0f %.0f %.0f V1 %.1f %.1f %.1f\n",
				    d.Xb, d.Xt, d.Xi, d.Gb, fabs(d.Gt), fabs(d.Gi));
				break;
			case PENTODE:
				printf("XU1 Vp Vg 0 Vs %s\n", d.name);
				printf("V3 Vs 0 %.0f\n", d.Vg2[0]);
				printf("V2 Vp 0 100\n");
				printf("V1 Vg 0 0\n");
				printf(".dc V2 %.0f %.0f %.0f V1 %.1f %.1f %.1f\n",
				    d.Xb, d.Xt, d.Xi, d.Gb, d.Gt, d.Gi);
				break;
			default:
				errx(1, "unknown tube for spice test");
			}
			printf(".lib tubelib.sub\n");
			printf(".backanno\n");
			printf(".end\n");
			printf("*\n* Use -I(V2) as curve to display\n");
			printf("* Adjust the rule on the left side accordingly\n*\n");
		}
	}

	if (d.RGI == 0.0)
		d.RGI = 2000;

	if (sflag) {
		double kg2 = 0.0;
		FILE *ofn = stdout;

		if (tflag) {
			char buf[1000];
			sprintf(buf, "%s.tsub", d.name);
			if ((ofn = fopen(buf, "w")) == NULL)
				err(1, "fopen");
		}

		/* Generate (lt)spice code */
		fprintf(ofn, "\n----------------------------- SPICE args ------------\n");
		fprintf(ofn, "*\n* %s code from %s generated by kurvmatch\n*\n", d.name, d.descr);
		switch (d.type) {
		case DIODE:
			fprintf(ofn, ".SUBCKT %s 1 2\n", d.name);
			fprintf(ofn, "XZ1 1 2 Diode CCP=%.1fP KA=%f EX=%.2f KB=%f EPS=%f\n",
			    d.CCP, dx[xKA], dx[xEX], dx[xKB], d.EPS);
			fprintf(ofn, ".ENDS\n");
			break;
		case PENTODE:
			/* Generate KG2 value */
			for (i = 0; i < d.sn; i++)
				kg2 += exp(dx[xEX] * log(d.Vs[i]/dx[xMU]))/d.Is[i];
			kg2 /= d.sn;
			fprintf(ofn, ".SUBCKT %s 1 2 3 4\n", d.name);
			fprintf(ofn, "XZ1 1 2 3 4 Pentode CCG=%.1fP CGP=%.1fP CCP=%.1fP "
			    "MU=%.0f EX=%.2f KG1=%.0f KP=%.0f KVB=%.0f RGI=%.0f KG2=%.0f\n",
			    d.CCG, d.CGP, d.CCP, dx[xMU], dx[xEX], dx[xKG1], dx[xKP],
			    dx[xKVB], d.RGI, kg2);
			fprintf(ofn, ".ENDS\n");
			break;
		case TRIODE:
			fprintf(ofn, ".SUBCKT %s 1 2 3\n", d.name);
			fprintf(ofn, "XZ1 1 2 3 Triode CCG=%.1fP CGP=%.1fP CCP=%.1fP "
			    "MU=%.0f EX=%.2f KG1=%.0f KP=%.0f KVB=%.0f RGI=%.0f\n",
			    d.CCG, d.CGP, d.CCP, dx[xMU], dx[xEX], dx[xKG1], dx[xKP], dx[xKVB], d.RGI);
			fprintf(ofn, ".ENDS\n");
			break;
		default:
			errx(1, "unknown spice type");
		}
		if (tflag)
			fclose(ofn);
	}

	return 0;
}
